@@ -2,12 +2,17 @@
 """
 Pandas like dataset API
 """
+import os
 import warnings
 import requests
 import pandas as pd
+import threading
+import time
+import uuid
+import concurrent.futures
 
 from pathlib import Path
-from typing import Callable, Dict, List, Optional, Union
+from typing import Any, Callable, Dict, List, Optional, Union
 from tqdm.auto import tqdm
 
 from relevanceai.dataset.read import Read
@@ -16,8 +21,9 @@ from relevanceai.utils import DocUtils
 from relevanceai.utils.logger import FileLogger
 from relevanceai.utils.decorators.analytics import track
 from relevanceai.utils import make_id
-
+from relevanceai.utils import fire_and_forget
 from relevanceai.constants.warning import Warning
+from relevanceai.utils.progress_bar import progress_bar
 
 
 class Write(Read):
@@ -29,12 +35,16 @@ class Write(Read):
         self,
         documents: list,
         bulk_fn: Callable = None,
-        max_workers: int = 8,
+        max_workers: Optional[int] = 2,
+        media_workers: Optional[int] = None,
         retry_chunk_mult: float = 0.5,
         show_progress_bar: bool = False,
         chunksize: int = 0,
         use_json_encoder: bool = True,
         create_id: bool = False,
+        overwrite: bool = True,
+        ingest_in_background: bool = False,
+        media_fields: Optional[List[str]] = None,
         **kwargs,
     ) -> Dict:
 
@@ -62,6 +72,8 @@ class Write(Read):
             Number of documents to upload per worker. If None, it will default to the size specified in config.upload.target_chunk_mb
         use_json_encoder : bool
             Whether to automatically convert documents to json encodable format
+        media_fields: List[str]
+            specifies which fields are local medias and need to upserted to S3. These should be given in absolute path format
 
         Example
         --------
@@ -88,6 +100,13 @@ class Write(Read):
             df.insert_documents(documents)
 
         """
+        if media_fields is not None:
+            documents = self.prepare_media_documents(
+                documents,
+                media_fields,
+                max_workers=media_workers,
+            )
+
         results = self._insert_documents(
             dataset_id=self.dataset_id,
             documents=documents,
@@ -98,6 +117,8 @@ class Write(Read):
             chunksize=chunksize,
             use_json_encoder=use_json_encoder,
             create_id=create_id,
+            overwrite=overwrite,
+            ingest_in_background=ingest_in_background,
             **kwargs,
         )
         return self._process_insert_results(results)
@@ -107,7 +128,7 @@ class Write(Read):
         self,
         filepath_or_buffer,
         chunksize: int = 10000,
-        max_workers: int = 8,
+        max_workers: int = 2,
         retry_chunk_mult: float = 0.5,
         show_progress_bar: bool = False,
         index_col: int = None,
@@ -208,6 +229,8 @@ class Write(Read):
         self.print_search_dashboard_url(self.dataset_id)
         return results
 
+    insert_df = insert_pandas_dataframe
+
     @track
     def insert_media_folder(
         self,
@@ -305,13 +328,14 @@ class Write(Read):
         self,
         documents: list,
         bulk_fn: Callable = None,
-        max_workers: int = 8,
+        max_workers: int = 2,
         retry_chunk_mult: float = 0.5,
         chunksize: int = 0,
         show_progress_bar=False,
         use_json_encoder: bool = True,
         return_json: bool = False,
         create_id: bool = False,
+        ingest_in_background: bool = False,
     ) -> Dict:
 
         """
@@ -370,6 +394,7 @@ class Write(Read):
             chunksize=chunksize,
             use_json_encoder=use_json_encoder,
             create_id=create_id,
+            ingest_in_background=ingest_in_background,
         )
         return self._process_insert_results(results, return_json=return_json)
 
@@ -475,9 +500,7 @@ class Write(Read):
         retrieve_chunksize: int = 100,
         filters: Optional[list] = None,
         select_fields: Optional[list] = None,
-        show_progress_bar: bool = True,
-        use_json_encoder: bool = True,
-        log_to_file: bool = True,
+        max_active_threads: int = 2,
     ):
         """
         Apply a bulk function along an axis of the DataFrame.
@@ -516,19 +539,32 @@ class Write(Read):
 
             df.apply(update_documents)
         """
+        thread_count = 0
         filters = [] if filters is None else filters
         select_fields = [] if select_fields is None else select_fields
 
-        return self.pull_update_push_async(
-            self.dataset_id,
-            bulk_func,
-            retrieve_chunk_size=retrieve_chunksize,
-            filters=filters,
+        for chunk in self.chunk_dataset(
             select_fields=select_fields,
-            show_progress_bar=show_progress_bar,
-            use_json_encoder=use_json_encoder,
-            log_to_file=log_to_file,
-        )
+            filters=filters,
+            chunksize=retrieve_chunksize,
+        ):
+            updated_chunk = bulk_func(
+                chunk,
+            )
+
+            @fire_and_forget
+            def fire_upsert_docs():
+                self.upsert_documents(updated_chunk)
+
+            thread_count += 1
+            if thread_count >= max_active_threads:
+                # Check if thread count decreases
+                curr_thread_count = threading.active_count()
+                while threading.active_count() >= curr_thread_count:
+                    time.sleep(1)
+                thread_count -= 1
+
+            fire_upsert_docs()
 
     @track
     def cat(self, vector_name: Union[str, None] = None, fields: Optional[List] = None):
@@ -610,7 +646,7 @@ class Write(Read):
     def create(self, schema: Optional[dict] = None) -> Dict:
         """
         A dataset can store documents to be searched, retrieved, filtered and aggregated (similar to Collections in MongoDB, Tables in SQL, Indexes in ElasticSearch).
-        A powerful and core feature of VecDB is that you can store both your metadata and vectors in the same document. When specifying the schema of a dataset and inserting your own vector use the suffix (ends with) "_vector_" for the field name, and specify the length of the vector in dataset_schema. \n
+        A powerful and core feature of Relevance is that you can store both your metadata and vectors in the same document. When specifying the schema of a dataset and inserting your own vector use the suffix (ends with) "_vector_" for the field name, and specify the length of the vector in dataset_schema. \n
 
         For example:
 
@@ -634,7 +670,7 @@ class Write(Read):
                 "product_text_chunkvector_" : 1024
             }
 
-        You don't have to specify the schema of every single field when creating a dataset, as VecDB will automatically detect the appropriate data type for each field (vectors will be automatically identified by its "_vector_" suffix). Infact you also don't always have to use this endpoint to create a dataset as /datasets/bulk_insert will infer and create the dataset and schema as you insert new documents. \n
+        You don't have to specify the schema of every single field when creating a dataset, as Relevance will automatically detect the appropriate data type for each field (vectors will be automatically identified by its "_vector_" suffix). Infact you also don't always have to use this endpoint to create a dataset as /datasets/bulk_insert will infer and create the dataset and schema as you insert new documents. \n
 
         Note:
 
@@ -696,7 +732,6 @@ class Write(Read):
         """
         return self.datasets.delete(self.dataset_id)
 
-    insert_df = insert_pandas_dataframe
 
     def _upload_media(
         self, presigned_url: str, media_content: bytes, verbose: bool = True
@@ -710,6 +745,23 @@ class Write(Read):
             if verbose:
                 print("media successfully uploaded.")
 
+    @track
+    def insert_media_bytes(self, bytes: bytes, filename: str, verbose: bool = True):
+        """
+        Insert a single media URL
+        """
+        # media to download
+        response = self.datasets.get_file_upload_urls(self.dataset_id, files=[filename])
+        url = response["files"][0]["url"]
+        self._upload_media(
+            presigned_url=response["files"][0]["upload_url"],
+            media_content=bytes,
+            verbose=verbose,
+        )
+        if verbose:
+            print(f"media is hosted at {url}")
+        return url
+
     @track
     def insert_media_url(self, media_url: str, verbose: bool = True):
         """
@@ -734,6 +786,7 @@ class Write(Read):
         media_urls: List[str],
         verbose: bool = True,
         file_log: str = "insert_media_urls.log",
+        logging: bool = True,
     ):
         """
         Insert a single media URL
@@ -741,9 +794,28 @@ class Write(Read):
         # media to download
         response = self.datasets.get_file_upload_urls(self.dataset_id, files=media_urls)
         response_docs: dict = {"media_documents": [], "failed_medias": []}
-        with FileLogger(file_log):
+        if logging:
+            with FileLogger(file_log):
+                for i, im in enumerate(tqdm(media_urls)):
+                    response_doc = {"_id": str(uuid.uuid4())}
+                    response_doc["media_file"] = im
+                    response_doc["media_url"] = response["files"][i]["url"]
+                    try:
+                        self._upload_media(
+                            presigned_url=response["files"][i]["upload_url"],
+                            media_content=requests.get(im).content,
+                            verbose=verbose,
+                        )
+                        response_docs["media_documents"].append(response_doc)
+                    except Exception as e:
+                        if verbose:
+                            print(f"Failed to upload {im}.")
+                        if verbose:
+                            print(e)
+                        response_docs["failed_medias"].append(response_doc)
+        else:
             for i, im in enumerate(tqdm(media_urls)):
-                response_doc = {}
+                response_doc = {"_id": str(uuid.uuid4())}
                 response_doc["media_file"] = im
                 response_doc["media_url"] = response["files"][i]["url"]
                 try:
@@ -798,6 +870,7 @@ class Write(Read):
         media_fns: List[str],
         verbose: bool = False,
         file_log="local_media_upload.log",
+        logging: bool = True,
     ):
         """Insert a list of local medias.
 
@@ -812,9 +885,27 @@ class Write(Read):
         """
         response = self.datasets.get_file_upload_urls(self.dataset_id, files=media_fns)
         response_docs: dict = {"media_documents": [], "failed_medias": []}
-        with FileLogger(file_log) as f:
+
+        if logging:
+            with FileLogger(file_log) as f:
+                for i, media_fn in enumerate(tqdm(media_fns)):
+                    response_doc = {"_id": str(uuid.uuid4())}
+                    response_doc["media_file"] = media_fn
+                    response_doc["media_url"] = response["files"][i]["url"]
+                    try:
+                        self._upload_media(
+                            presigned_url=response["files"][i]["upload_url"],
+                            media_content=self._open_local_media(media_fn),
+                            verbose=verbose,
+                        )
+                        response_docs["media_documents"].append(response_doc)
+                    except Exception as e:
+                        print(f"failed to upload {media_fn}")
+                        print(e)
+                        response_docs["failed_medias"].append(response_doc)
+        else:
             for i, media_fn in enumerate(tqdm(media_fns)):
-                response_doc = {}
+                response_doc = {"_id": str(uuid.uuid4())}
                 response_doc["media_file"] = media_fn
                 response_doc["media_url"] = response["files"][i]["url"]
                 try:
@@ -828,6 +919,7 @@ class Write(Read):
                     print(f"failed to upload {media_fn}")
                     print(e)
                     response_docs["failed_medias"].append(response_doc)
+
         return response_docs
 
     def get_media_documents(
@@ -835,6 +927,7 @@ class Write(Read):
         media_fns: List[str],
         verbose: bool = False,
         file_log: str = "media_upload.log",
+        logging: bool = True,
     ) -> dict:
         """
         Bulk insert medias. Returns a link to once it has been hosted
@@ -850,18 +943,29 @@ class Write(Read):
         """
         # Algorithm aims to insert local or hosted medias
         if "http" in media_fns[0]:
-            return self.insert_media_urls(media_fns, verbose=verbose, file_log=file_log)
+            return self.insert_media_urls(
+                media_fns,
+                verbose=verbose,
+                file_log=file_log,
+                logging=logging,
+            )
         else:
             return self.insert_local_medias(
-                media_fns, verbose=verbose, file_log=file_log
+                media_fns,
+                verbose=verbose,
+                file_log=file_log,
+                logging=logging,
             )
 
+    host_media_documents = get_media_documents
+
     @track
     def upsert_media(
         self,
         media_fns: List[str],
         verbose: bool = False,
         file_log: str = "media_upload.log",
+        logging: bool = True,
         **kw,
     ):
         """
@@ -878,7 +982,10 @@ class Write(Read):
             The file log to write
         """
         documents = self.get_media_documents(
-            media_fns=media_fns, verbose=verbose, file_log=file_log
+            media_fns=media_fns,
+            verbose=verbose,
+            file_log=file_log,
+            logging=logging,
         )
         return self.upsert_documents(documents["media_documents"], create_id=True, **kw)
 
@@ -918,3 +1025,148 @@ class Write(Read):
         return self.datasets.documents.update_where(
             dataset_id=self.dataset_id, update=update, filters=filters
         )
+
+    def insert_list(self, labels: list, label_field: str = "label", **kwargs):
+        """It takes a list of labels, and inserts them into the database as documents
+
+        Parameters
+        ----------
+        labels : list
+            list of labels to insert
+        label_field : str, optional
+            The field in the document that contains the label.
+
+        Returns
+        -------
+            A list of the ids of the documents that were inserted.
+
+        """
+        documents = [{label_field: l} for l in labels]
+        return self.insert_documents(documents=documents, **kwargs)
+
+    def batched_upsert_media(
+        self,
+        images: List[str],
+        show_progress_bar: bool = False,
+        n_workers: Optional[int] = None,
+    ) -> List[str]:
+        """
+        It takes a list of images, splits it into batches, and then uses a thread pool to upsert the
+        images in parallel
+
+        Parameters
+        ----------
+        images : List[str]
+            A list of media src paths to upload
+        show_progress_bar : bool
+            Show the progress bar
+        max_workers : Optional[int]
+            The number of workers to use. If None, this is set to the max number in ThreadPoolExecutor
+
+        Returns
+        -------
+            List[str]: A list of media_urls
+
+        """
+
+        if n_workers is None:
+            max_workers = os.cpu_count() + 4  # type: ignore
+        else:
+            max_workers = n_workers
+
+        bs = int(len(images) / max_workers)
+        nb = int(len(images) / bs)
+
+        with concurrent.futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
+            futures = []
+            for i in range(int(nb)):
+                futures.append(
+                    executor.submit(
+                        self.upsert_media,
+                        images[i * bs : (i + 1) * bs],
+                        False,
+                        "",
+                        False,
+                    )
+                )
+
+            for future in progress_bar(
+                concurrent.futures.as_completed(futures),
+                show_progress_bar=show_progress_bar,
+            ):
+                res = future.result()
+
+        return [document["media_url"] for document in self.get_all_documents()]
+
+    def prepare_media_documents(
+        self,
+        documents: List[Dict[str, Any]],
+        media_fields: List[str],
+        max_workers: Optional[int] = None,
+    ) -> List[Dict[str, Any]]:
+
+        list_of_media_url_mappings = []
+
+        for media_field in media_fields:
+            paths = [document[media_field] for document in documents]
+            flat_paths = []
+            for path in paths:
+                if isinstance(path, str):
+                    flat_paths.append(path)
+                elif isinstance(path, list):
+                    flat_paths += path
+            paths = list(set(flat_paths))
+
+            def upload_media(path, url):
+
+                response_doc = {"_id": str(uuid.uuid4())}
+                response_doc["media_file"] = path
+
+                with open(path, "rb") as f:
+                    img = bytes(f.read())
+
+                requests.put(
+                    url["upload_url"],
+                    data=img,
+                )
+
+                del img
+
+                return url["url"]
+
+            urls = self.datasets.get_file_upload_urls(self.dataset_id, files=paths)[
+                "files"
+            ]
+
+            def upload():
+                with concurrent.futures.ThreadPoolExecutor(
+                    max_workers=max_workers
+                ) as executor:
+                    data = list(
+                        tqdm(executor.map(upload_media, paths, urls), total=len(paths))
+                    )
+                return data
+
+            url_mapping = {path: url for path, url in zip(paths, upload())}
+
+            list_of_media_url_mappings.append(url_mapping)
+
+        for media_field, media_url_mapping in zip(
+            media_fields, list_of_media_url_mappings
+        ):
+            for document in documents:
+                if isinstance(document[media_field], str):
+                    document[f"{media_field}_url"] = media_url_mapping[
+                        document[media_field]
+                    ]
+                    document[media_field] = os.path.split(document[media_field])[-1]
+
+                elif isinstance(document[media_field], list):
+                    document[f"{media_field}_url"] = [
+                        media_url_mapping[media] for media in document[media_field]
+                    ]
+                    document[media_field] = [
+                        os.path.split(media)[-1] for media in document[media_field]
+                    ]
+
+        return documents
