@@ -1,21 +1,21 @@
 import warnings
-
+import numpy as np
+import pandas as pd
 from copy import deepcopy
 
 from typing import Optional, Union, Callable, Dict, Any, Set, List
 
 from relevanceai.utils.decorators.analytics import track
 
-from relevanceai.operations_new.apibase import OperationAPIBase
-from relevanceai.operations_new.cluster.alias import ClusterAlias
-from relevanceai.operations_new.cluster.base import ClusterBase
+from relevanceai.operations_new.ops_base import OperationAPIBase
+from relevanceai.operations_new.cluster.transform import ClusterTransform
 
 from relevanceai.constants import Warning
 from relevanceai.constants.errors import MissingClusterError
 from relevanceai.constants import MissingClusterError, Warning
 
 
-class ClusterOps(ClusterBase, OperationAPIBase, ClusterAlias):
+class ClusterOps(ClusterTransform, OperationAPIBase):
     """
     Cluster-related functionalities
     """
@@ -32,7 +32,7 @@ class ClusterOps(ClusterBase, OperationAPIBase, ClusterAlias):
         verbose: bool = False,
         model=None,
         model_kwargs=None,
-        *args,
+        byo_cluster_field: str = None,
         **kwargs,
     ):
         """
@@ -68,59 +68,97 @@ class ClusterOps(ClusterBase, OperationAPIBase, ClusterAlias):
         )
 
         # alias is set after model so that we can get the number of clusters
-        # if the model needs ot be instantiated
+        # if the model needs to be instantiated
         self.alias = self._get_alias(alias)
 
-    def _operate(self, cluster_id: str, field: str, output: dict, func: Callable):
+        self.byo_cluster_field = byo_cluster_field
+        if byo_cluster_field is not None:
+            self.create_byo_clusters()
+
+    def insert_centroids(
+        self,
+        centroid_documents,
+    ) -> None:
         """
-        Internal function for operations
+        Insert centroids
+        Centroids look below
 
-        It takes a cluster_id, a field, an output dictionary, and a function, and then it gets all the
-        documents in the cluster, gets the field across all the documents, and then applies the function
-        to the field
+        .. code-block::
 
-        Parameters
-        ----------
-        cluster_id : str
-            str, field: str, output: dict, func: Callable
-        field : str
-            the field you want to get the value for
-        output : dict
-            dict
-        func : Callable
-            Callable
+            cluster_ops = client.ClusterOps(
+                vector_field=["sample_1_vector_"],
+                alias="sample"
+            )
+            cluster_ops.insert_centroids(
+                centorid_documents=[
+                    {"_id" : "cluster-0", "sample_1_vector_": [1, 1, 1]},
+                    {"_id" : "cluster-1", "sample_1_vector_": [1, 2, 2]},
+                ]
+            )
 
         """
-        cluster_field = self._get_cluster_field_name()
-        # TODO; change this to fetch all documents
-        documents = self._get_all_documents(
-            self.dataset_id,
-            filters=[
-                {
-                    "field": cluster_field,
-                    "filter_type": "exact_match",
-                    "condition": "==",
-                    "condition_value": cluster_id,
-                },
-                {
-                    "field": field,
-                    "filter_type": "exists",
-                    "condition": ">=",
-                    "condition_value": " ",
-                },
-            ],
-            select_fields=[field, cluster_field],
-            show_progress_bar=False,
+        # Centroid documents are in the format {"cluster-0": [1, 1, 1]}
+        return self.datasets.cluster.centroids.insert(
+            dataset_id=self.dataset_id,
+            cluster_centers=self.json_encoder(centroid_documents),
+            vector_fields=self.vector_fields,
+            alias=self.alias,
         )
-        # get the field across each
-        arr = self.get_field_across_documents(field, documents)
-        output[cluster_id] = func(arr)
 
-    def _operate_across_clusters(self, field: str, func: Callable):
-        output: Dict[str, Any] = dict()
-        for cluster_id in self.list_cluster_ids():
-            self._operate(cluster_id=cluster_id, field=field, output=output, func=func)
-        return output
+    def calculate_centroids(self):
+
+        # calculate the centroids
+        centroid_vectors = {}
+
+        def calculate_centroid(vectors):
+            X = np.array(vectors)
+            return X.mean(axis=0)
+
+        centroid_vectors = self._operate_across_clusters(
+            field=self.vector_fields[0], func=calculate_centroid
+        )
+
+        # Does this insert properly?
+        if isinstance(centroid_vectors, dict):
+            centroid_vectors = [
+                {"_id": k, self.vector_fields[0]: v}
+                for k, v in centroid_vectors.items()
+            ]
+        return centroid_vectors
+
+    def create_centroids(self, insert: bool = True):
+        """
+        Calculate centroids from your dataset vectors.
+
+        Example
+        --------
+
+        .. code-block::
+
+            from relevanceai import Client
+            client = Client()
+            ds = client.Dataset("sample")
+            cluster_ops = ds.ClusterOps(
+                alias="kmeans-25",
+                vector_fields=['sample_vector_']
+            )
+            centroids = cluster_ops.create_centroids()
+
+        """
+        # Get an array of the different vectors
+        if len(self.vector_fields) > 1:
+            raise NotImplementedError(
+                "Do not currently support multiple vector fields for centroid creation."
+            )
+
+        # calculate the centroids
+        centroid_vectors = self.calculate_centroids()
+
+        if insert:
+            self.insert_centroids(
+                centroid_documents=centroid_vectors,
+            )
+        return centroid_vectors
 
     def list_cluster_ids(
         self,
@@ -153,6 +191,8 @@ class ClusterOps(ClusterBase, OperationAPIBase, ClusterAlias):
             The number of clusters
 
         """
+        if alias is None:
+            alias = self.alias
         # Mainly to be used for subclustering
         # Get the cluster alias
         cluster_field = self._get_cluster_field_name()
@@ -184,94 +224,9 @@ class ClusterOps(ClusterBase, OperationAPIBase, ClusterAlias):
 
         return list(all_cluster_ids)
 
-    def insert_centroids(
-        self,
-        centroid_documents,
-    ) -> None:
-        """
-        Insert centroids
-        Centroids look below
-
-        .. code-block::
-
-            cluster_ops = client.ClusterOps(
-                vector_field=["sample_1_vector_"],
-                alias="sample"
-            )
-            cluster_ops.insert_centroids(
-                centorid_documents={
-                    "cluster-0": [1, 1, 1],
-                    "cluster-2": [2, 1, 1]
-                }
-            )
-
-        """
-        # Centroid documents are in the format {"cluster-0": [1, 1, 1]}
-        return self.datasets.cluster.centroids.insert(
-            dataset_id=self.dataset_id,
-            cluster_centers=self.json_encoder(centroid_documents),
-            vector_fields=self.vector_fields,
-            alias=self.alias,
-        )
-
-    def create_centroids(self):
-        """
-        Calculate centroids from your vectors
-
-        Example
-        --------
-
-        .. code-block::
-
-            from relevanceai import Client
-            client = Client()
-            ds = client.Dataset("sample")
-            cluster_ops = ds.ClusterOps(
-                alias="kmeans-25",
-                vector_fields=['sample_vector_']
-            )
-            centroids = cluster_ops.create_centroids()
-
-        """
-        # Get an array of the different vectors
-        if len(self.vector_fields) > 1:
-            raise NotImplementedError(
-                "Do not currently support multiple vector fields for centroid creation."
-            )
-
-        # calculate the centroids
-        centroid_vectors = self.calculate_centroids()
-
-        self.insert_centroids(
-            centroid_documents=centroid_vectors,
-        )
-        return centroid_vectors
-
-    def calculate_centroids(self):
-        import numpy as np
-
-        # calculate the centroids
-        centroid_vectors = {}
-
-        def calculate_centroid(vectors):
-            X = np.array(vectors)
-            return X.mean(axis=0)
-
-        centroid_vectors = self._operate_across_clusters(
-            field=self.vector_fields[0], func=calculate_centroid
-        )
-
-        # Does this insert properly?
-        if isinstance(centroid_vectors, dict):
-            centroid_vectors = [
-                {"_id": k, self.vector_fields[0]: v}
-                for k, v in centroid_vectors.items()
-            ]
-        return centroid_vectors
-
     def get_centroid_documents(self):
         centroid_vectors = {}
-        if self.model._centroids is not None:
+        if hasattr(self.model, "_centroids") and self.model._centroids is not None:
             centroid_vectors = self.model._centroids
             # get the cluster label function
             labels = range(len(centroid_vectors))
@@ -286,9 +241,56 @@ class ClusterOps(ClusterBase, OperationAPIBase, ClusterAlias):
             ]
         else:
             centroids = self.create_centroids()
-
         return centroids
 
+    @property
+    def centroids(self):
+        """
+        Access the centroids of your dataset easily
+
+        .. code-block::
+
+            ds = client.Dataset("sample")
+            cluster_ops = ds.ClusterOps(
+                vector_fields=["sample_vector_"],
+                alias="simple"
+            )
+            cluster_ops.centroids
+
+        """
+        if not hasattr(self, "_centroids"):
+            self._centroids = self.datasets.cluster.centroids.documents(
+                dataset_id=self.dataset_id,
+                vector_fields=self.vector_fields,
+                alias=self.alias,
+                page_size=9999,
+                include_vector=True,
+            )["results"]
+        return self._centroids
+
+    def get_centroid_from_id(
+        self,
+        cluster_id: str,
+    ) -> Dict[str, Any]:
+        """> It takes a cluster id and returns the centroid with that id
+
+        Parameters
+        ----------
+        cluster_id : str
+            The id of the cluster to get the centroid for.
+
+        Returns
+        -------
+            The centroid with the given id.
+
+        """
+
+        for centroid in self.centroids:
+            if centroid["_id"] == cluster_id:
+                return centroid
+
+        raise ValueError(f"Missing the centroid with id {cluster_id}")
+
     def list_closest(
         self,
         cluster_ids: Optional[list] = None,
@@ -367,7 +369,6 @@ class ClusterOps(ClusterBase, OperationAPIBase, ClusterAlias):
             verbose=verbose,
         )
 
-    @track
     def list_furthest(
         self,
         cluster_ids: Optional[List] = None,
@@ -478,9 +479,11 @@ class ClusterOps(ClusterBase, OperationAPIBase, ClusterAlias):
         """
         if isinstance(encode_fn_or_model, str):
             # Get the model
-            from relevanceai.operations_new.vectorize.text.base import VectorizeTextBase
+            from relevanceai.operations_new.vectorize.text.transform import (
+                VectorizeTextTransform,
+            )
 
-            self.model = VectorizeTextBase._get_model(encode_fn_or_model)
+            self.model = VectorizeTextTransform._get_model(encode_fn_or_model)
             encode_fn = self.model.encode
         else:
             encode_fn = encode_fn_or_model
@@ -525,54 +528,6 @@ class ClusterOps(ClusterBase, OperationAPIBase, ClusterAlias):
             ),
         )
 
-    @property
-    def centroids(self):
-        """
-        Access the centroids of your dataset easily
-
-        .. code-block::
-
-            ds = client.Dataset("sample")
-            cluster_ops = ds.ClusterOps(
-                vector_fields=["sample_vector_"],
-                alias="simple"
-            )
-            cluster_ops.centroids
-
-        """
-        if not hasattr(self, "_centroids"):
-            self._centroids = self.datasets.cluster.centroids.documents(
-                dataset_id=self.dataset_id,
-                vector_fields=self.vector_fields,
-                alias=self.alias,
-                page_size=9999,
-                include_vector=True,
-            )["results"]
-        return self._centroids
-
-    def get_centroid_from_id(
-        self,
-        cluster_id: str,
-    ) -> Dict[str, Any]:
-        """> It takes a cluster id and returns the centroid with that id
-
-        Parameters
-        ----------
-        cluster_id : str
-            The id of the cluster to get the centroid for.
-
-        Returns
-        -------
-            The centroid with the given id.
-
-        """
-
-        for centroid in self.centroids:
-            if centroid["_id"] == cluster_id:
-                return centroid
-
-        raise ValueError(f"Missing the centorid with id {cluster_id}")
-
     @staticmethod
     def _get_filters(
         filters: List[Dict[str, Union[str, int]]],
@@ -614,14 +569,248 @@ class ClusterOps(ClusterBase, OperationAPIBase, ClusterAlias):
 
         return filters
 
-    def merge(self, cluster_ids: list):
+    def merge(self, target_cluster_id: str, cluster_ids: list):
         """
-        Merge clusters into the first one.
+        Merge clusters into the target cluster.
         The centroids are re-calculated and become a new middle.
         """
         return self.datasets.cluster.merge(
             dataset_id=self.dataset_id,
             vector_fields=self.vector_fields,
             alias=self.alias,
-            cluster_ids=cluster_ids,
+            cluster_ids=[target_cluster_id] + cluster_ids,
+        )
+
+    def create_byo_clusters(self):
+        """
+        Create BYO clusters for a given field
+        """
+        # TODO: Change into generator to make unique values more than 9999
+        results = self.datasets.facets(
+            dataset_id=self.dataset_id, fields=[self.byo_cluster_field], page_size=9999
+        )
+
+        try:
+            for r in results["results"][self.byo_cluster_field]:
+                filters = [
+                    {
+                        "field": self.byo_cluster_field,
+                        "filter_type": "exact_match",
+                        "condition": "==",
+                        "condition_value": r["value"],
+                    }
+                ]
+                cluster_doc = {}
+                self.set_field(self._get_cluster_field_name(), cluster_doc, r["value"])
+                results = self.datasets.documents.update_where(
+                    dataset_id=self.dataset_id, update=cluster_doc, filters=filters
+                )
+
+        except KeyError:
+            raise ValueError("Cluster field has no values.")
+
+        return results
+
+    def _operate(self, cluster_id: str, field: str, output: dict, func: Callable):
+        """
+        Internal function for operations
+
+        It takes a cluster_id, a field, an output dictionary, and a function, and then it gets all the
+        documents in the cluster, gets the field across all the documents, and then applies the function
+        to the field
+
+        Parameters
+        ----------
+        cluster_id : str
+            str, field: str, output: dict, func: Callable
+        field : str
+            the field you want to get the value for
+        output : dict
+            dict
+        func : Callable
+            Callable
+
+        """
+        cluster_field = self._get_cluster_field_name()
+        # TODO; change this to fetch all documents
+        documents = self._get_all_documents(
+            self.dataset_id,
+            filters=[
+                {
+                    "field": cluster_field,
+                    "filter_type": "exact_match",
+                    "condition": "==",
+                    "condition_value": cluster_id,
+                },
+                {
+                    "field": field,
+                    "filter_type": "exists",
+                    "condition": ">=",
+                    "condition_value": " ",
+                },
+            ],
+            select_fields=[field, cluster_field],
+            show_progress_bar=False,
+        )
+        # get the field across each
+        arr = self.get_field_across_documents(field, documents)
+        output[cluster_id] = func(arr)
+
+    def _operate_across_clusters(self, field: str, func: Callable):
+        output: Dict[str, Any] = dict()
+        for cluster_id in self.list_cluster_ids():
+            self._operate(cluster_id=cluster_id, field=field, output=output, func=func)
+        return output
+
+    @property
+    def labels(self):
+        metadata = self.datasets.metadata(self.dataset_id)
+        metadata = metadata.get("results", {}).get("cluster_metadata", {})
+        vector_fields = ["text_mpnet_vector_"]
+        alias = "kmeans-100"
+
+        cluster_field = "_cluster_." + ".".join(vector_fields) + "." + alias
+        labels = metadata.get("labels", {}).get(cluster_field, {})
+        print("To view nicely, please use `pd.DataFrame(labels)`.")
+        return labels
+
+    @track
+    def aggregate(
+        self,
+        metrics: Optional[list] = None,
+        sort: Optional[list] = None,
+        groupby: Optional[list] = None,
+        filters: Optional[list] = None,
+        page_size: int = 20,
+        page: int = 1,
+        asc: bool = False,
+        flatten: bool = True,
+    ):
+        """
+        Takes an aggregation query and gets the aggregate of each cluster in a collection. This helps you interpret each cluster and what is in them.
+        It can only can be used after a vector field has been clustered. \n
+        Aggregation/Groupby of a collection using an aggregation query. The aggregation query is a json body that follows the schema of:
+
+        .. code-block::
+
+            {
+                "groupby" : [
+                    {"name": <alias>, "field": <field in the collection>, "agg": "category"},
+                    {"name": <alias>, "field": <another groupby field in the collection>, "agg": "numeric"}
+                ],
+                "metrics" : [
+                    {"name": <alias>, "field": <numeric field in the collection>, "agg": "avg"}
+                    {"name": <alias>, "field": <another numeric field in the collection>, "agg": "max"}
+                ]
+            }
+
+        For example, one can use the following aggregations to group score based on region and player name.
+
+        .. code-block::
+
+            {
+                "groupby" : [
+                    {"name": "region", "field": "player_region", "agg": "category"},
+                    {"name": "player_name", "field": "name", "agg": "category"}
+                ],
+                "metrics" : [
+                    {"name": "average_score", "field": "final_score", "agg": "avg"},
+                    {"name": "max_score", "field": "final_score", "agg": "max"},
+                    {'name':'total_score','field':"final_score", 'agg':'sum'},
+                    {'name':'average_deaths','field':"final_deaths", 'agg':'avg'},
+                    {'name':'highest_deaths','field':"final_deaths", 'agg':'max'},
+                ]
+            }
+        "groupby" is the fields you want to split the data into. These are the available groupby types:
+            - category : groupby a field that is a category
+            - numeric: groupby a field that is a numeric
+        "metrics" is the fields and metrics you want to calculate in each of those, every aggregation includes a frequency metric. These are the available metric types:
+            - "avg", "max", "min", "sum", "cardinality"
+        The response returned has the following in descending order. \n
+        If you want to return documents, specify a "group_size" parameter and a "select_fields" parameter if you want to limit the specific fields chosen. This looks as such:
+            .. code-block::
+                {
+                    'groupby':[
+                        {'name':'Manufacturer','field':'manufacturer','agg':'category',
+                        'group_size': 10, 'select_fields': ["name"]},
+                    ],
+                    'metrics':[
+                        {'name':'Price Average','field':'price','agg':'avg'},
+                    ],
+                }
+                # ouptut example:
+                {"title": {"title": "books", "frequency": 200, "documents": [{...}, {...}]}, {"title": "books", "frequency": 100, "documents": [{...}, {...}]}}
+        For array-aggregations, you can add "agg": "array" into the aggregation query.
+
+        Parameters
+        ----------
+        dataset_id : string
+            Unique name of dataset
+        metrics: list
+            Fields and metrics you want to calculate
+        groupby: list
+            Fields you want to split the data into
+        filters: list
+            Query for filtering the search results
+        page_size: int
+            Size of each page of results
+        page: int
+            Page of the results
+        asc: bool
+            Whether to sort results by ascending or descending order
+        flatten: bool
+            Whether to flatten
+        alias: string
+            Alias used to name a vector field. Belongs in field_{alias} vector
+        metrics: list
+            Fields and metrics you want to calculate
+        groupby: list
+            Fields you want to split the data into
+        filters: list
+            Query for filtering the search results
+        page_size: int
+            Size of each page of results
+        page: int
+            Page of the results
+        asc: bool
+            Whether to sort results by ascending or descending order
+        flatten: bool
+            Whether to flatten
+
+        Example
+        ---------
+        .. code-block::
+            from relevanceai import Client
+            client = Client()
+            df = client.Dataset("sample_dataset_id")
+            from sklearn.cluster import KMeans
+            model = KMeans(n_clusters=2)
+            cluster_ops = client.ClusterOps(alias="kmeans_2", model=model)
+            cluster_ops.run(df, vector_fields=["sample_vector_"])
+            clusterer.aggregate(
+                "sample_dataset_id",
+                groupby=[{
+                    "field": "title",
+                    "agg": "wordcloud",
+                }],
+                vector_fields=['sample_vector_']
+            )
+        """
+        metrics = [] if metrics is None else metrics
+        groupby = [] if groupby is None else groupby
+        filters = [] if filters is None else filters
+        sort = [] if sort is None else sort
+
+        return self.datasets.cluster.aggregate(
+            dataset_id=self.dataset_id,
+            vector_fields=self.vector_fields,
+            groupby=groupby,
+            metrics=metrics,
+            sort=sort,
+            filters=filters,
+            alias=self.alias,
+            page_size=page_size,
+            page=page,
+            asc=asc,
+            flatten=flatten,
         )
