@@ -1,9 +1,9 @@
 from typing import Any, List
 
 from relevanceai.operations_new.vectorize.ops import VectorizeOps
-from relevanceai.operations_new.vectorize.text.base import VectorizeTextBase
+from relevanceai.operations_new.vectorize.text.transform import VectorizeTextTransform
 
 
-class VectorizeTextOps(VectorizeOps, VectorizeTextBase):
+class VectorizeTextOps(VectorizeOps, VectorizeTextTransform):
     def __init__(self, *args, **kwargs):
         super().__init__(*args, **kwargs)
